module mod_equation_system

	use mod_parameters, only : dkind, solution_method,  n_fit, tn_fit, fixed_coords_option
	use mod_arrays, only : fixed_coords
!	use mod_transversal_commons, only : fixed_coords

	implicit none

	real(kind=dkind) :: theta_start = 0.d0
	! where to start from for points where conditions are enforced

	real(kind=dkind), dimension(:,:), allocatable :: mat_A
	! matrix of coefficients for the system

	real(kind=dkind), dimension(:), allocatable :: sys_rhs
	! right hand side of the system of equations

	integer, dimension(:), allocatable ::  i_enforce
	! index arrays of points used for rhs

	real(kind=dkind) :: time_start, time_end, time_diff_1, time_diff

	real(kind=dkind) :: Rlast, Zlast
	real(kind=dkind) :: fix_dist, fix_dist_sq !distance between fixed_coord points
	! used in set_fixed_coords_length and defined here for debugging reasons

	real(kind=dkind) :: boundary_len = 0.d0
	! length of boundary

	real(kind=dkind), dimension(:,:), allocatable :: len_of_theta
	! cumulative length as a function of theta

	real(kind=dkind) :: dth
	! delta theta, defined here to keep things clean

	contains

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine find_coil_currents
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! determines the coil currents using the method specified via input

! can calculate input plasma J for debugging
!	call checkJ

	if (solution_method==1) then
		call linear_system_solution
	elseif (solution_method==2) then
		call least_square_solution
	elseif (solution_method==3) then
		call least_square_solution_simple
	elseif (solution_method==4) then
		call least_square_solution_simple_with_gradients
	elseif (solution_method==5) then
		call linear_system_solution_with_gradients
	elseif (solution_method==11) then
		call least_square_solution_weighted
	elseif (solution_method==12) then
		call least_square_solution_weighted_with_gradients
	endif

	continue

end subroutine find_coil_currents


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine linear_system_solution
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!-----------LINEAR SYSTEM-----------
! builds and solves the linear system

	use mod_parameters, only : n_enforce
	use mod_coils, only : I_coil, I_CS

	real(kind=dkind), dimension(:,:), allocatable :: mat_A_copy
	real(kind=dkind), dimension(:), allocatable :: sys_rhs_copy
	! for debugging purposes
	integer, dimension(:), allocatable :: indx
	real(kind=dkind) :: d !this is apparently useless
	integer :: i, j

	call linear_system_setup

	allocate(indx(n_enforce))
	allocate(sys_rhs_copy(n_enforce))
	allocate(mat_A_copy(1:n_enforce,1:n_enforce))

	do j = 1, n_enforce
	do i = 1, n_enforce

		mat_A_copy(i,j) = mat_A(i,j)

	enddo
	enddo

	do i = 1, n_enforce

		sys_rhs_copy(i) = sys_rhs(i)

	enddo

	call ludcmp(mat_A_copy,n_enforce,n_enforce,indx,d)
	call lubksb(mat_A_copy,n_enforce,n_enforce,indx,sys_rhs_copy)

	do i = 1, n_enforce-1

		I_coil(i) = sys_rhs_copy(i)

	enddo

	I_CS = sys_rhs_copy(i)

	continue

	return

end subroutine linear_system_solution

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine linear_system_setup
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!-----------LINEAR SYSTEM-----------
! fills the coordinates of fixed_coords, enters the appropriate values in sys_rhs and mat_A

	use mod_parameters, only : n_enforce

	integer :: i

!!$	if(fixed_coords_option<10) then
!!$		fixed_coords_option = 2
!!$	endif

	call set_fixed_coords_wrapper

	call fill_mat_A

	allocate(sys_rhs(n_enforce))

	do i = 1, n_enforce

		call getrhs(fixed_coords(i,1), fixed_coords(i,2), sys_rhs(i))

	enddo

	continue

	return

end subroutine linear_system_setup

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine linear_system_solution_with_gradients
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!-----------LINEAR SYSTEM-----------
! builds and solves the linear system, using gradients on boundary.

	use mod_parameters, only : n_enforce
	use mod_coils, only : I_coil, I_CS

	real(kind=dkind), dimension(:,:), allocatable :: mat_A_copy
	real(kind=dkind), dimension(:), allocatable :: sys_rhs_copy
	! for debugging purposes
	integer, dimension(:), allocatable :: indx
	real(kind=dkind) :: d !this is apparently useless
	integer :: i, j

!	if(fixed_coords_option<10) then
!		fixed_coords_option = 2
!	endif

	call linear_system_with_gradients_setup

	allocate(indx(n_enforce))
	allocate(sys_rhs_copy(n_enforce))
	allocate(mat_A_copy(1:n_enforce,1:n_enforce))

	do j = 1, n_enforce
	do i = 1, n_enforce

		mat_A_copy(i,j) = mat_A(i,j)

	enddo
	enddo

	do i = 1, n_enforce

		sys_rhs_copy(i) = sys_rhs(i)

	enddo

	call ludcmp(mat_A_copy,n_enforce,n_enforce,indx,d)
	call lubksb(mat_A_copy,n_enforce,n_enforce,indx,sys_rhs_copy)

	do i = 1, n_enforce-1

		I_coil(i) = sys_rhs_copy(i)

	enddo

	I_CS = sys_rhs_copy(i)

	continue

	return

end subroutine linear_system_solution_with_gradients


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine linear_system_with_gradients_setup
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!-----------LINEAR SYSTEM-----------
! fills the coordinates of fixed_coords, enters the appropriate values in sys_rhs and mat_A

	use mod_parameters, only : n_enforce

	integer :: i, iloc1, iloc2

	call set_fixed_coords_wrapper
!	call set_fixed_coords
!	call set_fixed_coords3

	call fill_mat_A_with_gradients

	allocate(sys_rhs(n_enforce))

!	do i = 1, n_enforce

!		call getrhs(fixed_coords(i,1), fixed_coords(i,2), sys_rhs(i))

!	enddo

	do i = 1,n_enforce/2 ! n_fit

		iloc1 = 2*i-1
		iloc2 = 2*i

!		ind_list(iloc1) = 1.d0*iloc1
!		ind_list(iloc2) = 1.d0*iloc2

		call getrhs_with_gradients(fixed_coords(iloc1,1), fixed_coords(iloc1,2), sys_rhs(iloc1:iloc2),  &
														fixed_coords(iloc1,3), fixed_coords(iloc1,4))
!		sys_rhs(i) = -sys_rhs(i) "+" is the consistent sign

	enddo

	continue

	return

end subroutine linear_system_with_gradients_setup


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine set_fixed_coords_wrapper
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! decides how to assign the points where the plasma edge condition psi=0 is enforced.

	if(fixed_coords_option==1) then
		call set_fixed_coords
	elseif(fixed_coords_option==10) then
		call set_fixed_coords_length
	elseif(fixed_coords_option==11) then
		call set_fixed_coords_equal_arc
	else
		print*, 'error in fixed_coords_option:'
		print*, 'fixed_coords_option = ', fixed_coords_option
		print*, 'this option does not exist'
		pause
		stop
	endif

	continue

end subroutine set_fixed_coords_wrapper


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine set_fixed_coords
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! simplest approach, use an angle determined from the Cartesian coordinates R and Z and set points at constant intervals.

	use mod_parameters, only : pi, n_enforce
	use mod_coils, only : n_currents
	use mod_plasma, only : radius_theta, radius_theta_with_ders

	real(kind=dkind) :: theta_loc, r, Rloc, Zloc, Rprim, Zprim, dummy
	integer :: n_size, istep

	integer :: i

	n_enforce = n_currents + 1
	! the "+1" is for the central solenoid

	if(solution_method==1) then
		n_size = n_enforce
		n_fit = n_size
		istep = 1
	elseif(solution_method==2) then
		n_size = n_fit
		theta_start = 0.d0
		istep = 1
		! no need to start anywhere else than the origin
	elseif(solution_method==3) then
		n_size = n_fit
		theta_start = 0.d0
		istep = 1
		! no need to start anywhere else than the origin
	elseif(solution_method==4) then
		tn_fit = n_fit*2
		theta_start = 0.d0
		istep = 2
		n_size = tn_fit
		! no need to start anywhere else than the origin
		! two conditions imposed in each point
	elseif(solution_method==5) then
		n_size = n_enforce
		n_fit = n_size
		istep = 2
		! simplest approach: repeat points twice as in solution_method 4
	elseif(solution_method==11) then
		n_size = n_fit
		theta_start = 0.d0
		istep = 1
		! no need to start anywhere else than the origin
	elseif(solution_method==12) then
		tn_fit = n_fit*2
		theta_start = 0.d0
		istep = 2
		n_size = tn_fit
		! no need to start anywhere else than the origin
		! two conditions imposed in each point
	elseif(solution_method==101) then
		n_size = n_fit
		istep = 1
	endif

	allocate(fixed_coords(n_size,-1:4))
	! last two values are used only for solution method 4 and 5

	do i = 1, n_size, istep

		theta_loc = theta_start + 2.d0*pi*(i-1.d0)/n_size

		if((solution_method==4).or.(solution_method==5).or.(solution_method==12)) then
			call radius_theta_with_ders(theta_loc,r,Rloc,Zloc,dummy,Rprim,Zprim)
		else
			call radius_theta(theta_loc,r,Rloc, Zloc)
		endif

		fixed_coords(i,-1) = r
		fixed_coords(i,0) = theta_loc
		fixed_coords(i,1) = Rloc
		fixed_coords(i,2) = Zloc

		if((solution_method==4).or.(solution_method==5).or.(solution_method==12)) then
			fixed_coords(i,3) = Rprim
			fixed_coords(i,4) = Zprim
		endif

	enddo

	if((solution_method==4).or.(solution_method==5).or.(solution_method==12)) then

!		do i = 2, n_size-modulo(n_size,2), istep
!		! this takes care of an odd number of points for solution_method 5
		do i = 2, n_size, istep

			fixed_coords(i,-1) = fixed_coords(i-1,-1)
			fixed_coords(i,0) = fixed_coords(i-1,0) + 4.d0*pi
			fixed_coords(i,1) = fixed_coords(i-1,1)
			fixed_coords(i,2) = fixed_coords(i-1,2)
			fixed_coords(i,3) = fixed_coords(i-1,3)
			fixed_coords(i,4) = fixed_coords(i-1,4)

		enddo

	endif

	continue

	return

end subroutine set_fixed_coords


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine set_fixed_coords_length
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! (approximately)  calculates the length of the boundary, then distributes points at equal distances
! computationally expensive!

	use mod_parameters, only : pi, n_enforce
	use mod_transversal_commons, only : secroot
	use mod_coils, only : n_currents
	use mod_plasma, only : radius_theta, radius_theta_with_ders

	real(kind=dkind) :: theta_loc, r, Rloc, Zloc, Rprim, Zprim, dummy

	real(kind=dkind) :: theta_L, theta_R, theta_R_fix, dth_for_search !for search
    integer :: itemax = 200
    real(kind=dkind) :: xtoll = 1.d-6
    real(kind=dkind) :: ftoll = 1.d-6
	integer :: error

	logical :: repeat
	real(kind=dkind) :: last_dist
	real(kind=dkind) :: dist_tol = 1.d-3
	integer, parameter :: n_for_len = 10000
	integer :: n_size, istep

	integer :: i

	! first get a reasonable approximation of the length of the boundary

	call  compute_boundary_len(n_for_len)

	! now proceed with determining the points

	n_enforce = n_currents + 1
	! the "+1" is for the central solenoid

	if(solution_method==1) then
		n_size = n_enforce
		n_fit = n_size
		istep = 1
	elseif(solution_method==2) then
		n_size = n_fit
		theta_start = 0.d0
		istep = 1
		! no need to start anywhere else than the origin
	elseif(solution_method==3) then
		n_size = n_fit
		theta_start = 0.d0
		istep = 1
		! no need to start anywhere else than the origin
	elseif(solution_method==4) then
		tn_fit = n_fit*2
		theta_start = 0.d0
		istep = 2
		n_size = tn_fit
		! no need to start anywhere else than the origin
		! two conditions imposed in each point
	elseif(solution_method==5) then
		n_size = n_enforce
		n_fit = n_size
		istep = 2
		! simplest approach: repeat points twice as in solution_method 4
	elseif(solution_method==11) then
		n_size = n_fit
		theta_start = 0.d0
		istep = 1
		! no need to start anywhere else than the origin
	elseif(solution_method==12) then
		tn_fit = n_fit*2
		theta_start = 0.d0
		istep = 2
		n_size = tn_fit
		! no need to start anywhere else than the origin
		! two conditions imposed in each point
	endif

	allocate(fixed_coords(n_size,-1:4))
	! last two values are used only for solution method 4 and 5

	fix_dist = boundary_len/n_size * istep
	fix_dist_sq = fix_dist**2
!	dth_for_search = 3.d0*(2.d0*pi/n_size)
	dth_for_search = 1.5d0*(2.d0*pi/n_size)

	repeat = .true.

	do while(repeat)

		do i = 1, n_size, istep

			if(i==1) then

				theta_loc = 0.d0

			else

				theta_L = theta_loc
				theta_R_fix = theta_loc !+ dth_for_search
				error = 1

				do while(error>0)

					theta_R = theta_R_fix
					call secroot(dist_fun,theta_R,theta_L,theta_loc,itemax,xtoll,ftoll,error)
					theta_R_fix = theta_R_fix + 0.1d0*dth_for_search

				enddo

			endif

			if((solution_method==4).or.(solution_method==5).or.(solution_method==12)) then
				call radius_theta_with_ders(theta_loc,r,Rloc,Zloc,dummy,Rprim,Zprim)
			else
				call radius_theta(theta_loc,r,Rloc, Zloc)
			endif

			fixed_coords(i,-1) = r
			fixed_coords(i,0) = theta_loc
			fixed_coords(i,1) = Rloc
			fixed_coords(i,2) = Zloc

			if((solution_method==4).or.(solution_method==5).or.(solution_method==12)) then
				fixed_coords(i,3) = Rprim
				fixed_coords(i,4) = Zprim
			endif

			Rlast = Rloc
			Zlast = Zloc

		enddo

		last_dist = sqrt((Rloc-fixed_coords(1,1))**2+(Zloc-fixed_coords(1,2))**2)

		if(abs(last_dist/fix_dist-1.d0)/fix_dist>dist_tol) then

!			if(last_dist>fix_dist) then

				fix_dist = fix_dist + 0.1751d0*(last_dist-fix_dist)/ (n_size/istep + 1.d0)

!			fix_dist = (fix_dist*n_size/istep + last_dist) / (n_size/istep + 1.d0)
			fix_dist_sq = fix_dist**2

		else

			repeat = .false.

		endif

	enddo

	if((solution_method==4).or.(solution_method==5).or.(solution_method==12)) then

!		do i = 2, n_size-modulo(n_size,2), istep
!		! this takes care of an odd number of points for solution_method 5
		do i = 2, n_size, istep

			fixed_coords(i,-1) = fixed_coords(i-1,-1)
			fixed_coords(i,0) = fixed_coords(i-1,0) + 4.d0*pi
			fixed_coords(i,1) = fixed_coords(i-1,1)
			fixed_coords(i,2) = fixed_coords(i-1,2)
			fixed_coords(i,3) = fixed_coords(i-1,3)
			fixed_coords(i,4) = fixed_coords(i-1,4)

		enddo

	endif

	continue

	return

end subroutine set_fixed_coords_length

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
function dist_fun(theta_in) result(answer)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! return distance from input point to last recorded point

	use mod_plasma, only : radius_theta

	real(kind=dkind) :: theta_in, answer
	real(kind=dkind) :: r, Rloc, Zloc, dist_sq

	call radius_theta(theta_in,r,Rloc, Zloc)
	dist_sq = (Rloc-Rlast)**2 + (Zloc-Zlast)**2

	answer = dist_sq - fix_dist_sq

	continue

end function dist_fun


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine set_fixed_coords_equal_arc
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! (approximately)  calculates the length of the boundary, then distributes points at equal arc-length distances
! computationally expensive!

	use mod_parameters, only : pi, n_enforce
	use mod_transversal_commons, only : secroot
	use mod_coils, only : n_currents
	use mod_plasma, only : radius_theta, radius_theta_with_ders, theta_points

	real(kind=dkind) :: theta_loc, r, Rloc, Zloc, Rprim, Zprim, dummy

	real(kind=dkind) :: theta_L, theta_R !for search
    real(kind=dkind) :: xtoll = 1.d-6
    real(kind=dkind) :: ftoll = 1.d-6
	integer :: error

	logical :: repeat
!	real(kind=dkind) :: last_dist
!	real(kind=dkind) :: dist_tol = 1.d-3
	real(kind=dkind) :: arc_len, len_i
	integer :: n_for_len
	integer :: n_size, istep
	integer :: ilast

	integer :: i

	! first get a reasonable approximation of the length of the boundary

	n_for_len = max(theta_points,10*n_fit)
	! number of points to calculate boundary length
 
	call compute_boundary_len(n_for_len)

	! now proceed with determining the points

	n_enforce = n_currents + 1
	! the "+1" is for the central solenoid

	if(solution_method==1) then
		n_size = n_enforce
		n_fit = n_size
		istep = 1
	elseif(solution_method==2) then
		n_size = n_fit
		theta_start = 0.d0
		istep = 1
		! no need to start anywhere else than the origin
	elseif(solution_method==3) then
		n_size = n_fit
		theta_start = 0.d0
		istep = 1
		! no need to start anywhere else than the origin
	elseif(solution_method==4) then
		tn_fit = n_fit*2
		theta_start = 0.d0
		istep = 2
		n_size = tn_fit
		! no need to start anywhere else than the origin
		! two conditions imposed in each point
	elseif(solution_method==5) then
		n_size = n_enforce
		n_fit = n_size
		istep = 2
		! simplest approach: repeat points twice as in solution_method 4
	elseif(solution_method==11) then
		n_size = n_fit
		theta_start = 0.d0
		istep = 1
		! no need to start anywhere else than the origin
	elseif(solution_method==12) then
		tn_fit = n_fit*2
		theta_start = 0.d0
		istep = 2
		n_size = tn_fit
		! no need to start anywhere else than the origin
		! two conditions imposed in each point
	endif

	allocate(fixed_coords(n_size,-1:4))
	! last two values are used only for solution method 4 and 5

	arc_len = boundary_len/(n_size/istep)
	ilast = 1

	do i = 1, n_size, istep

		if(i==1) then

			theta_loc = 0.d0

		else

			repeat = .true.
			len_i = arc_len*(i-1)/istep

			do while(repeat)

				ilast = ilast+1
				if(len_of_theta(ilast,2)>len_i) then
					repeat = .false.
				endif

			enddo

			! just use linear interpolation for now
			theta_loc = len_of_theta(ilast-1,1) + dth*(len_i-len_of_theta(ilast-1,2)) /  &
							(len_of_theta(ilast,2)-len_of_theta(ilast-1,2))

		endif

		if((solution_method==4).or.(solution_method==5).or.(solution_method==12)) then
			call radius_theta_with_ders(theta_loc,r,Rloc,Zloc,dummy,Rprim,Zprim)
		else
			call radius_theta(theta_loc,r,Rloc, Zloc)
		endif

		fixed_coords(i,-1) = r
		fixed_coords(i,0) = theta_loc
		fixed_coords(i,1) = Rloc
		fixed_coords(i,2) = Zloc

		if((solution_method==4).or.(solution_method==5).or.(solution_method==12)) then
			fixed_coords(i,3) = Rprim
			fixed_coords(i,4) = Zprim
		endif

		Rlast = Rloc
		Zlast = Zloc

	enddo


	if((solution_method==4).or.(solution_method==5).or.(solution_method==12)) then

!		do i = 2, n_size-modulo(n_size,2), istep
!		! this takes care of an odd number of points for solution_method 5
		do i = 2, n_size, istep

			fixed_coords(i,-1) = fixed_coords(i-1,-1)
			fixed_coords(i,0) = fixed_coords(i-1,0) + 4.d0*pi
			fixed_coords(i,1) = fixed_coords(i-1,1)
			fixed_coords(i,2) = fixed_coords(i-1,2)
			fixed_coords(i,3) = fixed_coords(i-1,3)
			fixed_coords(i,4) = fixed_coords(i-1,4)

		enddo

	endif

	continue

	return

end subroutine set_fixed_coords_equal_arc



!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine compute_boundary_len(n_for_len)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

	use mod_parameters, only : pi
	use mod_plasma, only : radius_theta_with_ders

	integer :: n_for_len
	real(kind=dkind) :: theta_loc, th_m_half, dth_half
	real(kind=dkind) :: r,Rloc,Zloc,rprim,bigRprim,bigZprim
	integer :: i

	allocate(len_of_theta(n_for_len,2))
	! cumulative length as a function of theta

	theta_loc = 0.d0
	dth = 2.d0*pi/(n_for_len-1.d0)
	dth_half = dth/2.d0
	th_m_half = -dth_half

	len_of_theta(1,1) = theta_loc
	len_of_theta(1,2) = boundary_len

	do i = 1, n_for_len-2

		theta_loc = theta_loc + dth
		th_m_half = th_m_half + dth

		call radius_theta_with_ders(th_m_half,r,Rloc,Zloc,rprim,bigRprim,bigZprim)

		boundary_len = boundary_len + dth *  r * sqrt(1.d0+(rprim/r)**2)

		len_of_theta(i+1,1) = theta_loc
		len_of_theta(i+1,2) = boundary_len

	enddo

	! do last point separately for convenience
		call radius_theta_with_ders(-dth_half,r,Rloc,Zloc,rprim,bigRprim,bigZprim)

		boundary_len = boundary_len + dth *  r * sqrt(1.d0+(rprim/r)**2)

		continue

end subroutine compute_boundary_len

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine getrhs(Rrhs, Zrhs, rhs_loc)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! builds right hand side of linear system (direct version)

	use mod_plasma, only : nx, nz, dx, dz, dx_a, dz_a, x_coord, z_coord, sort_grid, j_phi

	implicit none

	real (kind=dkind), intent(in) :: Rrhs, Zrhs
	real (kind=dkind) :: rhs_loc
	real(kind=dkind) :: Green_loc, term_loc
	real (kind=dkind) :: dA
	real (kind=dkind) :: bigR, bigZ

!	real (kind=dkind) :: Green
!	external Green

	integer :: i,j

!	real(kind=dkind) :: R_l, R_r, angle
!	real(kind=dkind) :: psi_integral(2)

	! first, get the current and the cross section area

	rhs_loc = 0.d0

!	dA = dx*dz

	do j=1, nz
		do i=1, nx

		    if(sort_grid(i,j,0)>=1) then
			! may modify to >1 only

				bigR = x_coord(i) + dx_a(i) -  dx_a(i-1)
				bigZ = z_coord(j) + dz_a(j) -  dz_a(j-1)

!				Green_loc = Green(bigR,bigZ,Rrhs,Zrhs)
				call Green_single_term(bigR,bigZ,Rrhs,Zrhs,Green_loc)

				dA = 0.25d0*(dx_a(i) + dx_a(i-1)) * (dz_a(j) + dz_a(j-1))

				term_loc = j_phi(i,j) * Green_loc * dA
				rhs_loc = rhs_loc + term_loc

			endif

		enddo
	enddo

	rhs_loc = -rhs_loc

	continue

	return

end subroutine getrhs


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine getrhs_with_gradients(Rrhs, Zrhs, rhs_loc,Rprim,Zprim)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! builds right hand side of linear system (including gradients)

	use mod_plasma, only : nx, nz, dx, dz, dx_a, dz_a, x_coord, z_coord, sort_grid, j_phi

	implicit none

	real (kind=dkind), intent(in) :: Rrhs, Zrhs, Rprim, Zprim
	real (kind=dkind) :: rhs_loc(2)
	real(kind=dkind) :: term_loc, Green_loc, Green_R_loc, Green_Z_loc
	real (kind=dkind) :: dA
	real (kind=dkind) :: bigR, bigZ

	integer :: i,j

!	real(kind=dkind) :: R_l, R_r, angle
!	real(kind=dkind) :: psi_integral(2)

	! first, get the current and the cross section area

	rhs_loc = 0.d0

!	dA = dx*dz

	do j=1, nz
		do i=1, nx

		    if(sort_grid(i,j,0)>=1) then
			! may modify to >1 only

				bigR = x_coord(i) + dx_a(i) -  dx_a(i-1)
				bigZ = z_coord(j) + dz_a(j) -  dz_a(j-1)

				call Green_terms(bigR,bigZ,Rrhs,Zrhs,Green_loc,Green_R_loc,Green_Z_loc)

				dA = 0.25d0*(dx_a(i) + dx_a(i-1)) * (dz_a(j) + dz_a(j-1))

				term_loc = j_phi(i,j) * Green_loc * dA
				rhs_loc(1) = rhs_loc(1) + term_loc

!				term_loc = j_phi(i,j) * dA * (Green_R_loc*Zprim - Green_Z_loc*Rprim)
				term_loc = j_phi(i,j) * dA * (Green_R_loc*Rprim + Green_Z_loc*Zprim)
				rhs_loc(2) = rhs_loc(2) + term_loc

			endif

		enddo
	enddo

	rhs_loc = -rhs_loc

	continue

	return

end subroutine getrhs_with_gradients


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine fill_mat_A
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! mat_A is filled this way:
! each row corresponds to a point where conditions are enforced;
! since the vector of unknowns is the vector of coil currents (CS last),
! each element ij  is the contribution of coil j to condition in point i.

	use mod_parameters, only : n_enforce
	use mod_coils, only : R_coil, Z_coil, repeat_index

	real (kind=dkind) :: Green
	external Green

	real(kind=dkind) :: Rp, Zp, Rc, Zc
	integer :: i, j, k

	allocate(mat_A(1:n_enforce,1:n_enforce))
	mat_A = 0.d0

	do i = 1, n_enforce
	do j = 1, n_enforce-1
	do k = 1, repeat_index(j)

		Rp = fixed_coords(i,1)
		Zp = fixed_coords(i,2)

		Rc = R_coil(j,k)
		Zc = Z_coil(j,k)

		mat_A(i,j) = mat_A(i,j) + Green(Rp, Zp, Rc, Zc)

	enddo
	enddo
	enddo

	call mat_A_CS_element

	continue

	return

end subroutine fill_mat_A

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine mat_A_CS_element
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

	use mod_parameters, only : n_enforce
	use mod_coils, only : n_CS, R_CS, Z_CS

	real (kind=dkind) :: Green
	external Green

	real(kind=dkind) :: Rp, Zp, Rc, Zc
	real(kind=dkind) :: term_loc
	integer :: i, j, k

	j = n_enforce

	do i = 1, n_enforce

		Rp = fixed_coords(i,1)
		Zp = fixed_coords(i,2)

		do k = 1, n_CS

			Rc = R_CS(k)
			Zc = Z_CS(k)

			term_loc = Green(Rp, Zp, Rc, Zc)
			mat_A(i,j) = mat_A(i,j) + term_loc

		enddo

	enddo

	continue

	return

end subroutine mat_A_CS_element

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine fill_CS_coordinates
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

	use mod_coils, only : n_CS, R_CS, Z_CS, CS_points, dR_CS,  &
						dZ_CS, dA_hat_CS, coord_CS_point

	real(kind=dkind) :: dx_loc, dz_loc, x_loc, z_loc
	integer :: nR_CS, nZ_CS, n_ratio
	integer :: i, j, k, p

	n_ratio = dZ_CS(1)/dR_CS(1)
	
	nR_CS = CS_points/(n_ratio+1)
	nR_CS = max(nR_CS,1)
	nZ_CS = nR_CS * n_ratio
	CS_points = nZ_CS * nR_CS
	
	dx_loc = dR_CS(1)/nR_CS
	dz_loc = dZ_CS(1)/nZ_CS

!	do p = 1, n_CS
	p = 1

	do i = 1, nR_CS

		x_loc = R_CS(p) + (i-0.5d0)*dx_loc

		do j = 1, nZ_CS
	
			z_loc = Z_CS(p) + (j-0.5d0)*dz_loc
			k = (i-1)*nZ_CS + j
			coord_CS_point(p,k,1) = x_loc
			coord_CS_point(p,k,2) = z_loc
	
		enddo

	enddo





	continue

	return

end subroutine fill_CS_coordinates


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine fill_mat_A_with_gradients
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! mat_A is filled this way:
! each row corresponds to a point where conditions are enforced;
! since the vector of unknowns is the vector of coil currents (CS last),
! each element ij  is the contribution of coil j to condition in point i.

	use mod_parameters, only : n_enforce
	use mod_coils, only : R_coil, Z_coil, repeat_index

	real (kind=dkind) :: Green_loc,Green_R_loc,Green_Z_loc
	real(kind=dkind) :: Rp, Zp, Rprim, Zprim, Rc, Zc
	integer :: i, j, k

	allocate(mat_A(1:n_enforce,1:n_enforce))
	mat_A = 0.d0

	do i = 1, n_enforce,2
	do j = 1, n_enforce-1
	do k = 1, repeat_index(j)

		Rp = fixed_coords(i,1)
		Zp = fixed_coords(i,2)
		Rprim = fixed_coords(i,3)
		Zprim = fixed_coords(i,4)

		Rc = R_coil(j,k)
		Zc = Z_coil(j,k)

		call Green_terms(Rc,Zc,Rp,Zp,Green_loc,Green_R_loc,Green_Z_loc)

		mat_A(i,j) = mat_A(i,j) + Green_loc

		if(i<n_enforce) then
			mat_A(i+1,j) = mat_A(i+1,j) + Green_R_loc*Rprim + Green_Z_loc*Zprim
		endif

	enddo
	enddo
	enddo

	call mat_A_with_gradients_CS_element

	continue

	return

end subroutine fill_mat_A_with_gradients

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine mat_A_with_gradients_CS_element
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

	use mod_parameters, only : n_enforce
	use mod_coils, only : n_CS, R_CS, Z_CS

	real (kind=dkind) :: Green_loc,Green_R_loc,Green_Z_loc
	real(kind=dkind) :: Rp, Zp, Rprim, Zprim, Rc, Zc

	integer :: i, j, k

	j = n_enforce

	do i = 1, n_enforce, 2

		Rp = fixed_coords(i,1)
		Zp = fixed_coords(i,2)
		Rprim = fixed_coords(i,3)
		Zprim = fixed_coords(i,4)

		do k = 1, n_CS

			Rc = R_CS(k)
			Zc = Z_CS(k)

			call Green_terms(Rc,Zc,Rp,Zp,Green_loc,Green_R_loc,Green_Z_loc)
			mat_A(i,j) = mat_A(i,j) + Green_loc

			if(i<n_enforce) then
				mat_A(i+1,j) = mat_A(i+1,j) + Green_R_loc*Rprim + Green_Z_loc*Zprim
			endif

		enddo

	enddo

	continue

	return

end subroutine mat_A_with_gradients_CS_element


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine least_square_solution
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! uses SVD decomposition to find optimal coil current values

	use mod_parameters, only : n_enforce
	use mod_coils, only : I_coil, I_CS

	real(kind=dkind) :: chisq
	real(kind=dkind), dimension(:,:), allocatable :: v_temp,u_temp
	real(kind=dkind), dimension(:), allocatable :: w_temp, solution_array, sigma, ind_list

	integer :: i

	call cpu_time(time_start)

	call set_fixed_coords_wrapper
	! call set_fixed_coords

	allocate(solution_array(1:n_enforce))
	allocate(sigma(1:n_fit))
	allocate(ind_list(1:n_fit))

	allocate(sys_rhs(n_fit))

	do i = 1, n_fit

		call getrhs(fixed_coords(i,1), fixed_coords(i,2), sys_rhs(i))
!		sys_rhs(i) = -sys_rhs(i) "+" is the consistent sign
		ind_list(i) = 1.d0*i

	enddo

	allocate(v_temp(1:n_enforce,1:n_enforce))
	allocate(u_temp(1:n_fit,1:n_fit))
	allocate(w_temp(1:n_enforce))

	sigma = 1.d2 ! equal weight for all points?
	solution_array = 1.d0 ! use all coils

	call cpu_time(time_end)
	time_diff_1 = time_end - time_start

	call svdfit(ind_list,sys_rhs,sigma,n_fit,solution_array,n_enforce,u_temp,v_temp,w_temp,n_fit,n_enforce,chisq,fill_element_for_svd)

	call cpu_time(time_end)

	time_diff = time_end - time_start

	open(69, file='computation time.txt')
	write(69,*) 'matrix assembly', time_diff_1
	write(69,*) 'total time', time_diff
	close(69)

	do i = 1, n_enforce-1

		I_coil(i) = solution_array(i)

	enddo

	I_CS = solution_array(i)

	continue

end subroutine least_square_solution


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine least_square_solution_simple
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! uses linear least square fit to find optimal coil current values
 

	use mod_parameters, only : n_enforce
	use mod_coils, only : I_coil, I_CS

	real(kind=dkind) :: chisq
	real(kind=dkind), dimension(:,:), allocatable :: v_temp,u_temp, covar
	real(kind=dkind), dimension(:), allocatable :: w_temp, solution_array, sigma, ind_list
	integer, dimension(:), allocatable :: ones

	integer :: i

	call cpu_time(time_start)

	call set_fixed_coords_wrapper
!	call set_fixed_coords

	allocate(solution_array(1:n_enforce))
	allocate(sigma(1:n_fit))
	allocate(ones(1:n_enforce))
	allocate(covar(1:n_enforce,1:n_enforce))
	allocate(ind_list(1:n_fit))

	allocate(sys_rhs(n_fit))

	do i = 1, n_fit

		call getrhs(fixed_coords(i,1), fixed_coords(i,2), sys_rhs(i))
!		sys_rhs(i) = -sys_rhs(i) "+" is the consistent sign
		ind_list(i) = 1.d0*i

	enddo

	allocate(v_temp(1:n_enforce,1:n_enforce))
	allocate(u_temp(1:n_fit,1:n_fit))
	allocate(w_temp(1:n_enforce))

	sigma = 1.d2 ! equal weight for all points?
	solution_array = 1.d0 ! use all coils

	do i =1, n_enforce
		ones(i) = i ! use all coils
	enddo

	call cpu_time(time_end)
	time_diff_1 = time_end - time_start

!	call svdfit(fixed_coords(:,0),sys_rhs,sigma,n_fit,solution_array,n_enforce,u_temp,v_temp,w_temp,n_fit,n_enforce,chisq,fill_element_for_svd)
	call  lfit(ind_list,sys_rhs,sigma,n_fit,solution_array,n_enforce,ones,n_enforce,covar,n_enforce,chisq,fill_element_for_svd)
!				(x                           ,y,           sig,      ndata,a,                       ma,        lista,mfit,covar,ncvm,chisq,funcs)

	call cpu_time(time_end)

	time_diff = time_end - time_start

	open(69, file='computation time.txt')
	write(69,*) 'matrix assembly', time_diff_1
	write(69,*) 'total time', time_diff
	close(69)

	do i = 1, n_enforce-1

		I_coil(i) = solution_array(i)

	enddo

	I_CS = solution_array(i)

	continue

end subroutine least_square_solution_simple


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine least_square_solution_simple_with_gradients
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! uses simple least square fit to find optimal coil current values

	use mod_parameters, only : n_enforce
	use mod_coils, only : I_coil, I_CS

	real(kind=dkind) :: chisq
	real(kind=dkind), dimension(:,:), allocatable :: v_temp,u_temp, covar
	real(kind=dkind), dimension(:), allocatable :: w_temp, solution_array, sigma, ind_list
	integer, dimension(:), allocatable :: ones

	integer :: i, iloc1, iloc2

	call cpu_time(time_start)

	call set_fixed_coords_wrapper
!	call set_fixed_coords

	allocate(solution_array(1:n_enforce))
	allocate(sigma(1:tn_fit))
	allocate(ones(1:n_enforce))
	allocate(covar(1:n_enforce,1:n_enforce))

	allocate(sys_rhs(tn_fit))
	allocate(ind_list(tn_fit))


	do i = 1, n_fit

		iloc1 = 2*i-1
		iloc2 = 2*i

		ind_list(iloc1) = 1.d0*iloc1
		ind_list(iloc2) = 1.d0*iloc2

		call getrhs_with_gradients(fixed_coords(iloc1,1), fixed_coords(iloc1,2), sys_rhs(iloc1:iloc2),  &
														fixed_coords(iloc1,3), fixed_coords(iloc1,4))
!		sys_rhs(i) = -sys_rhs(i) "+" is the consistent sign

	enddo

	allocate(v_temp(1:n_enforce,1:n_enforce))
	allocate(u_temp(1:tn_fit,1:tn_fit))
	allocate(w_temp(1:n_enforce))

	sigma = 1.d2 ! equal weight for all points?
	solution_array = 1.d0 ! use all coils

	do i =1, n_enforce
		ones(i) = i ! use all coils
	enddo

	call cpu_time(time_end)
	time_diff_1 = time_end - time_start

!	call svdfit(fixed_coords(:,0),sys_rhs,sigma,n_fit,solution_array,n_enforce,u_temp,v_temp,w_temp,n_fit,n_enforce,chisq,fill_element_for_svd)
	call  lfit(ind_list,sys_rhs,sigma,tn_fit,solution_array,n_enforce,ones,n_enforce,covar,n_enforce,chisq,fill_element_for_svd_with_grads)
!				(x                           ,y,           sig,      ndata,a,                       ma,        lista,mfit,covar,ncvm,chisq,funcs)

	call cpu_time(time_end)
	time_diff = time_end - time_start

	open(69, file='computation time.txt')
	write(69,*) 'matrix assembly', time_diff_1
	write(69,*) 'total time', time_diff
	close(69)

	do i = 1, n_enforce-1

		I_coil(i) = solution_array(i)

	enddo

	I_CS = solution_array(i)

	continue

end subroutine least_square_solution_simple_with_gradients

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine least_square_solution_weighted
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! This is very similar to its namesake (repeated only for clarity). The only differences are that LHS and RHS are
! divided by the equilibrium |grad psi| AND all equations are normalized to be non-dimensional

	use mod_parameters, only : n_enforce
	use mod_coils, only : I_coil, I_CS
	use mod_plasma, only : norm_grad
	use mod_transversal_commons, only : get_psi_gradient

	real(kind=dkind) :: chisq
	real(kind=dkind), dimension(:,:), allocatable :: v_temp,u_temp
	real(kind=dkind), dimension(:), allocatable :: w_temp, solution_array, sigma, ind_list

	integer :: i

	call cpu_time(time_start)

	call set_fixed_coords_wrapper
!	call set_fixed_coords

	allocate(solution_array(1:n_enforce))
	allocate(sigma(1:n_fit))
	allocate(ind_list(1:n_fit))

	allocate(sys_rhs(n_fit))

	do i = 1, n_fit

		call getrhs(fixed_coords(i,1), fixed_coords(i,2), sys_rhs(i))
!		sys_rhs(i) = -sys_rhs(i) "+" is the consistent sign
		ind_list(i) = 1.d0*i

	enddo

	call get_psi_gradient

	allocate(v_temp(1:n_enforce,1:n_enforce))
	allocate(u_temp(1:n_fit,1:n_fit))
	allocate(w_temp(1:n_enforce))

	do i = 1, n_fit

		sigma(i) = norm_grad(i)

	enddo

!	sigma = 1.d2 ! equal weight for all points?

	solution_array = 1.d0 ! use all coils

	call cpu_time(time_end)
	time_diff_1 = time_end - time_start

	call svdfit(ind_list,sys_rhs,sigma,n_fit,solution_array,n_enforce,u_temp,v_temp,w_temp,n_fit,n_enforce,chisq,fill_element_for_svd)

	call cpu_time(time_end)

	time_diff = time_end - time_start

	open(69, file='computation time.txt')
	write(69,*) 'matrix assembly', time_diff_1
	write(69,*) 'total time', time_diff
	close(69)

	do i = 1, n_enforce-1

		I_coil(i) = solution_array(i)

	enddo

	I_CS = solution_array(i)

	continue

end subroutine least_square_solution_weighted


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine least_square_solution_weighted_with_gradients
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! SVD least square solution with gradients, all equations are non-dimensional and normalized to the edge
! given |grad psi|

	use mod_parameters, only : n_enforce
	use mod_coils, only : I_coil, I_CS
	use mod_plasma, only : norm_grad
	use mod_transversal_commons, only : get_psi_gradient

	real(kind=dkind) :: chisq
	real(kind=dkind), dimension(:,:), allocatable :: v_temp,u_temp, covar
	real(kind=dkind), dimension(:), allocatable :: w_temp, solution_array, sigma, ind_list
	integer, dimension(:), allocatable :: ones

	integer :: i, iloc1, iloc2

	call cpu_time(time_start)

	call set_fixed_coords_wrapper
!	call set_fixed_coords

	allocate(solution_array(1:n_enforce))
	allocate(sigma(1:tn_fit))
	allocate(ones(1:n_enforce))
	allocate(covar(1:n_enforce,1:n_enforce))

	allocate(sys_rhs(tn_fit))
	allocate(ind_list(tn_fit))


	do i = 1, n_fit

		iloc1 = 2*i-1
		iloc2 = 2*i

		ind_list(iloc1) = 1.d0*iloc1
		ind_list(iloc2) = 1.d0*iloc2

		call getrhs_with_gradients(fixed_coords(iloc1,1), fixed_coords(iloc1,2), sys_rhs(iloc1:iloc2),  &
														fixed_coords(iloc1,3), fixed_coords(iloc1,4))
!		sys_rhs(i) = -sys_rhs(i) "+" is the consistent sign

	enddo

	call get_psi_gradient

	allocate(v_temp(1:n_enforce,1:n_enforce))
	allocate(u_temp(1:tn_fit,1:tn_fit))
	allocate(w_temp(1:n_enforce))

	do i = 1, n_fit

! original
!		sigma(2*i-1) = norm_grad(i) * fixed_coords(i,-1)
!		sigma(2*i) = norm_grad(i)
		sigma(2*i-1) = norm_grad(i)
		sigma(2*i) = norm_grad(i) * fixed_coords(i,-1)

	enddo

	solution_array = 1.d0 ! use all coils

	do i =1, n_enforce
		ones(i) = i ! use all coils
	enddo

	call cpu_time(time_end)
	time_diff_1 = time_end - time_start

!	call svdfit(fixed_coords(:,0),sys_rhs,sigma,n_fit,solution_array,n_enforce,u_temp,v_temp,w_temp,n_fit,n_enforce,chisq,fill_element_for_svd)
	call  lfit(ind_list,sys_rhs,sigma,tn_fit,solution_array,n_enforce,ones,n_enforce,covar,n_enforce,chisq,fill_element_for_svd_with_grads)
!				(x                           ,y,           sig,      ndata,a,                       ma,        lista,mfit,covar,ncvm,chisq,funcs)

	call cpu_time(time_end)
	time_diff = time_end - time_start

	open(69, file='computation time.txt')
	write(69,*) 'matrix assembly', time_diff_1
	write(69,*) 'total time', time_diff
	close(69)

	do i = 1, n_enforce-1

		I_coil(i) = solution_array(i)

	enddo

	I_CS = solution_array(i)

	continue

end subroutine least_square_solution_weighted_with_gradients


!The user supplies a subroutine funcs(x,afunc,ma) that returns the ma basis functions evaluated at x = x in the array afunc.
! basis function: Green function for each coil (CS included)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine fill_element_for_svd(ind_loc, coeffs_loc, n_useless)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

	use mod_parameters, only : n_enforce
	use mod_coils, only : R_coil, Z_coil, R_CS, Z_CS, n_CS, repeat_index
	use mod_plasma, only : radius_theta

	real(kind=dkind) :: ind_loc
	real(kind=dkind) :: coeffs_loc(n_enforce)
	integer :: n_useless	!to comply with library structure, this is n_enforce

!	real(kind=dkind) :: theta_loc
	real(kind=dkind) :: Rp, Zp, Rc, Zc, term_loc, rloc

	integer :: ind
	integer :: j, k, h

	real (kind=dkind) :: Green
	external Green

	ind = int(ind_loc)
!	theta_loc = fixed_coords(ind,0)
	Rp = fixed_coords(ind,1)
	Zp = fixed_coords(ind,2)

	coeffs_loc = 0.d0

	do j = 1, n_enforce-1
	do h = 1, repeat_index(j)

		Rc = R_coil(j,h)
		Zc = Z_coil(j,h)

		call Green_single_term(Rp, Zp, Rc, Zc,term_loc)
		coeffs_loc(j) = coeffs_loc(j) + term_loc
!		coeffs_loc(j) = coeffs_loc(j) + Green(Rp, Zp, Rc, Zc)

	enddo
	enddo

	do k = 1, n_CS

		Rc = R_CS(k)
		Zc = Z_CS(k)

!		term_loc = Green(Rp, Zp, Rc, Zc)
		call Green_single_term(Rp, Zp, Rc, Zc,term_loc)
		coeffs_loc(j) = coeffs_loc(j) + term_loc

	enddo


end subroutine fill_element_for_svd

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine fill_element_for_svd_with_grads(ind_loc, coeffs_loc, n_useless)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! this is called twice for each value of theta (second time with argument theta+2pi, second call
! immediately follows first one)
! this is done to include both conditions on the value of psi and on the gradient
! to take advantage of the way coefficients are calculated, both series of coefficients are calculated
! at the first call, then just moved to the output arrays at the second call
! this version uses an integer as input

	use mod_parameters, only : pi, n_enforce
	use mod_coils, only : R_coil, Z_coil, R_CS, Z_CS, n_CS, repeat_index
	use mod_plasma, only : radius_theta_with_ders, r_data, theta_points


	real(kind=dkind) :: ind_loc
	real(kind=dkind) :: theta_loc
	real(kind=dkind) :: coeffs_loc(n_enforce)
	integer :: n_useless	!to comply with library structure, this is n_enforce
	integer :: ind

	real(kind=dkind) :: Rp, Zp, Rc, Zc, term_loc, rloc, Green_loc, Green_R_loc, Green_Z_loc, Rprim, Zprim
	real(kind=dkind), dimension(:), allocatable, save :: coeffs_save
	integer, save :: ind_last

	integer :: j, k, h

	ind = int(ind_loc)
	theta_loc = fixed_coords(ind,0)

	if(allocated(coeffs_save)) then
		continue
	else
		allocate(coeffs_save(n_enforce))
!		ind_last = 0
	endif

	coeffs_loc = 0.d0

	if(theta_loc<=r_data(theta_points,1)) then
!	if(theta_loc<2.d0*pi) then
!	if(ind/=ind_last) then
	! calculate all coefficients

		! clear previous values
		coeffs_save = 0.d0

		Rp = fixed_coords(ind,1)
		Zp = fixed_coords(ind,2)
		Rprim = fixed_coords(ind,3)
		Zprim = fixed_coords(ind,4)

		do j = 1, n_enforce-1
		do h = 1, repeat_index(j)

			Rc = R_coil(j,h)
			Zc = Z_coil(j,h)

			call Green_terms(Rc,Zc,Rp,Zp,Green_loc,Green_R_loc,Green_Z_loc)

			coeffs_loc(j) = coeffs_loc(j) + Green_loc

!			coeffs_save(j) = coeffs_save(j) + Green_R_loc*Zprim - Green_Z_loc*Rprim
			coeffs_save(j) = coeffs_save(j) + Green_R_loc*Rprim + Green_Z_loc*Zprim

		enddo
		enddo

		do k = 1, n_CS

			Rc = R_CS(k)
			Zc = Z_CS(k)

			call Green_terms(Rc,Zc,Rp,Zp,Green_loc,Green_R_loc,Green_Z_loc)

			coeffs_loc(j) = coeffs_loc(j) + Green_loc

!			coeffs_save(j) = coeffs_save(j) + Green_R_loc*Zprim - Green_Z_loc*Rprim
			coeffs_save(j) = coeffs_save(j) + Green_R_loc*Rprim + Green_Z_loc*Zprim

		enddo

!		ind_last = ind

	else
	! coefficients were calculated at previous call, just pack them in

		do j = 1, n_enforce

			coeffs_loc(j) = coeffs_save(j)

		enddo

	endif

	continue

end subroutine fill_element_for_svd_with_grads




!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine ludcmp(a,n,np,indx,d)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! from numerical recipes

! Given a matrix a(1:n,1:n), with physical dimension np by np, this routine replaces it by
! the LU decomposition of a row-wise permutation of itself. a and n arei nput. a is output,
! arranged as in equation (2.3.14) above; indx(1:n) is an output vector that records the
! row permutation eﬀected by the partial pivoting; d is output as
! ±1depending on whether
! the number of row interchanges was even or odd, respectively. This routine is used in
! combination with lubksb to solve linear equations or invert a matrix.

	integer, parameter :: nmax=1000
	real(kind=dkind), parameter :: tiny=1.0d-20

	integer :: n, np
	real(kind=dkind) :: a(np,np),vv(nmax)
	integer :: indx(n)

	real(kind=dkind) :: d, aamax, sum, dum
	integer :: i,j,k, imax

	d=1.d0

	do i=1,n

		aamax=0.d0

		do j=1,n
			if (abs(a(i,j)).gt.aamax) aamax=abs(a(i,j))
		enddo

		if (aamax.eq.0.d0) pause 'singular matrix.'
		vv(i)=1.d0/aamax

	enddo

	do j=1,n

		if (j.gt.1) then

			do i=1,j-1

				sum=a(i,j)

				if (i.gt.1)then
					do k=1,i-1
						sum=sum-a(i,k)*a(k,j)
					enddo
					a(i,j)=sum
				endif

			enddo

		endif

		aamax=0.d0

		do i=j,n

			sum=a(i,j)

			if (j.gt.1)then

				do k=1,j-1
					sum=sum-a(i,k)*a(k,j)
				enddo

				a(i,j)=sum

			endif

			dum=vv(i)*abs(sum)

			if (dum.ge.aamax) then
				imax=i
				aamax=dum
			endif

		enddo

		if (j.ne.imax)then

			do k=1,n
				dum=a(imax,k)
				a(imax,k)=a(j,k)
				a(j,k)=dum
			enddo

			d=-d
			vv(imax)=vv(j)

		endif

		indx(j)=imax

		if(j.ne.n)then

			if(a(j,j).eq.0.d0)a(j,j)=tiny
			dum=1./a(j,j)

			do i=j+1,n
				a(i,j)=a(i,j)*dum
			enddo

		endif

	enddo

	if(a(n,n).eq.0.d0)a(n,n)=tiny

	return

end subroutine ludcmp

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine lubksb(a,n,np,indx,b)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! from numerical recipes

! Solves the set of n linear equationsA·X=B. Here a is input, not as the matrix A but
! rather as its LU decomposition, determined by the routine ludcmp. indx is input as the
! permutation vector returned by ludcmp. b(1:n) is input as the right-hand side vector B,
! and returns with the solution vector X. a, n, np, and indx are not modiﬁed by this routine
! and can be left in place for successive calls with diﬀerent right-hand sides b. This routine
! takes into account the possibility that b will beginwith many zero elements, so it is efficient
! for use in matrix inversion.

	integer :: n, np
	integer :: indx(n)
	real(kind=dkind) :: a(np,np),b(n)
	real(kind=dkind) :: sum
	integer :: i, j, ii, ll

	ii=0
	do 12 i=1,n
	ll=indx(i)
	sum=b(ll)
	b(ll)=b(i)
	if (ii.ne.0)then
	do 11 j=ii,i-1
	sum=sum-a(i,j)*b(j)
	11        continue
	else if (sum.ne.0.d0) then
	ii=i
	endif
	b(i)=sum
	12    continue
	do 14 i=n,1,-1
	sum=b(i)
	if(i.lt.n)then
	do 13 j=i+1,n
	sum=sum-a(i,j)*b(j)
	13        continue
	endif
	b(i)=sum/a(i,i)
	14    continue

return

end subroutine lubksb


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine svdfit(x,y,sig,ndata,a,ma,u,v,w,mp,np,chisq,funcs)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! from numerical recipes

	integer, parameter :: nmax=1000,mmax=50
	real(kind=dkind), parameter :: tol=1.d-5

	integer :: ndata, ma, mp, np
	real(kind=dkind):: x(ndata),y(ndata),sig(ndata),a(ma),v(np,np),u(mp,np),w(np),b(nmax),afunc(mmax)
	real(kind=dkind) :: wmax, chisq, sum, tmp, thresh, s, f

	external funcs

	integer :: i, j

	do i=1,ndata

		call funcs(x(i),afunc,ma)
		tmp=1.d0/sig(i)

		do j=1,ma
			u(i,j)=afunc(j)*tmp
		enddo

		b(i)=y(i)*tmp

	enddo

	call svdcmp(u,ndata,ma,mp,np,w,v)

	wmax=0.d0

	do j=1,ma
		if(w(j).gt.wmax)wmax=w(j)
	enddo

	thresh=tol*wmax

	do j=1,ma
		if(w(j).lt.thresh)w(j)=0.
	enddo

	call svbksb(u,w,v,ndata,ma,mp,np,b,a)
	chisq=0.d0

	do i=1,ndata

		call funcs(x(i),afunc,ma)
		sum=0.d0

		do j=1,ma
			sum=sum+a(j)*afunc(j)
		enddo

		chisq=chisq+((y(i)-sum)/sig(i))**2

	enddo

	return

end subroutine svdfit


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine svdcmp(a,m,n,mp,np,w,v)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! from numerical recipes

	integer, parameter :: nmax=100
	integer :: m,n,mp,np
	real(kind=dkind) :: a(mp,np),w(np),v(np,np),rv1(nmax)
	real(kind=dkind) :: g, scale, anorm, s, f, h, c, x, y, z

	integer :: i, j, l, k, its, jj
	integer :: nm

	g=0.0
	scale=0.0
	anorm=0.0

	do i=1,n

		l=i+1
		rv1(i)=scale*g
		g=0.0
		s=0.0
		scale=0.0

		if (i.le.m) then

			do k=i,m
				scale=scale+abs(a(k,i))
			enddo

			if (scale.ne.0.d0) then
	
				do k=i,m
					a(k,i)=a(k,i)/scale
					s=s+a(k,i)*a(k,i)
				enddo
		
				f=a(i,i)
				g=-sign(sqrt(s),f)
				h=f*g-s
				a(i,i)=f-g
		
				if (i.ne.n) then

					do j=l,n

						s=0.0

						do k=i,m
							s=s+a(k,i)*a(k,j)
						enddo

						f=s/h

						do k=i,m
							a(k,j)=a(k,j)+f*a(k,i)
						enddo

					enddo

				endif

			do k= i,m
				a(k,i)=scale*a(k,i)
			enddo

			endif
			endif

			w(i)=scale *g
			g=0.0
			s=0.0
			scale=0.0

		if ((i.le.m).and.(i.ne.n)) then

			do k=l,n
				scale=scale+abs(a(i,k))
			enddo

			if (scale.ne.0.0) then

				do k=l,n
					a(i,k)=a(i,k)/scale
					s=s+a(i,k)*a(i,k)
				enddo

				f=a(i,l)
				g=-sign(sqrt(s),f)
				h=f*g-s
				a(i,l)=f-g

				do k=l,n
					rv1(k)=a(i,k)/h
				enddo

				if (i.ne.m) then

					do j=l,m

						s=0.0

						do k=l,n
							s=s+a(j,k)*a(i,k)
						enddo

						do k=l,n
							a(j,k)=a(j,k)+s*rv1(k)
						enddo

					enddo

				endif

				do k=l,n
					a(i,k)=scale*a(i,k)
				enddo

			endif

		endif

		anorm=max(anorm,(abs(w(i))+abs(rv1(i))))

	enddo

	do i=n,1,-1

		if (i.lt.n) then

			if (g.ne.0.0) then

				do j=l,n
					v(j,i)=(a(i,j)/a(i,l))/g
				enddo


				do j=l,n

					s=0.0

					do k=l,n
						s=s+a(i,k)*v(k,j)
					enddo

					do k=l,n
						v(k,j)=v(k,j)+s*v(k,i)
					enddo

				enddo

			endif

			do j=l,n
				v(i,j)=0.0
				v(j,i)=0.0
			enddo

		endif

		v(i,i)=1.0
		g=rv1(i)
		l=i

	enddo

	do i=n,1,-1

		l=i+1
		g=w(i)

		if (i.lt.n) then

			do j=l,n
				a(i,j)=0.0
			enddo

		endif

		if (g.ne.0.0) then
	
			g=1.0/g

			if (i.ne.n) then

				do j=l,n

					s=0.0

					do k=l,m
						s=s+a(k,i)*a(k,j)
					enddo

					f=(s/a(i,i))*g

					do k=i,m
						a(k,j)=a(k,j)+f*a(k,i)
					enddo

				enddo

			endif

			do j=i,m
				a(j,i)=a(j,i)*g
			enddo

		else

			do j= i,m
				a(j,i)=0.0
			enddo

		endif

		a(i,i)=a(i,i)+1.0

	enddo

	do 49 k=n,1,-1

	do 48 its=1,30

	do 41 l=k,1,-1
		nm=l-1
		if ((abs(rv1(l))+anorm).eq.anorm)  go to 2
		if ((abs(w(nm))+anorm).eq.anorm)  go to 1
	41        continue

	1         c=0.0
	s=1.0

	do 43 i=l,k
		f=s*rv1(i)
		if ((abs(f)+anorm).ne.anorm) then
			g=w(i)
			h=sqrt(f*f+g*g)
			w(i)=h
			h=1.0/h
			c= (g*h)
			s=-(f*h)
			do 42 j=1,m
				y=a(j,nm)
				z=a(j,i)
				a(j,nm)=(y*c)+(z*s)
				a(j,i)=-(y*s)+(z*c)
			42            continue
		endif
	43        continue
	2         z=w(k)

	if (l.eq.k) then
	if (z.lt.0.0) then

	w(k)=-z

	do j=1,n
		v(j,k)=-v(j,k)
	enddo

	endif

	go to 3

	endif

	if (its.eq.30) pause 'no convergence in 30 iterations'

	x=w(l)
	nm=k-1
	y=w(nm)
	g=rv1(nm)
	h=rv1(k)
	f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y)
	g=sqrt(f*f+1.0)
	f=((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x
	c=1.0
	s=1.0

	do 47 j=l,nm

		i=j+1
		g=rv1(i)
		y=w(i)
		h=s*g
		g=c*g
		z=sqrt(f*f+h*h)
		rv1(j)=z
		c=f/z
		s=h/z
		f= (x*c)+(g*s)
		g=-(x*s)+(g*c)
		h=y*s
		y=y*c

		do 45 nm=1,n
			x=v(nm,j)
			z=v(nm,i)
			v(nm,j)= (x*c)+(z*s)
			v(nm,i)=-(x*s)+(z*c)
		45          continue

		z=sqrt(f*f+h*h)
		w(j)=z

		if (z.ne.0.0) then
			z=1.0/z
			c=f*z
			s=h*z
		endif

		f= (c*g)+(s*y)
		x=-(s*g)+(c*y)

		do 46 nm=1,m
			y=a(nm,j)
			z=a(nm,i)
			a(nm,j)= (y*c)+(z*s)
			a(nm,i)=-(y*s)+(z*c)
		46          continue

	47        continue

	rv1(l)=0.0
	rv1(k)=f
	w(k)=x
	48      continue
	3       continue
	49    continue

	return

end subroutine svdcmp


!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine svbksb(u,w,v,m,n,mp,np,b,x)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! from numerical recipes

	integer, parameter :: nmax=100
	integer :: m,n,mp,np
	real(kind=dkind) :: u(mp,np),w(np),v(np,np),b(mp),x(np),tmp(nmax)
	real(kind=dkind) :: s

	integer :: i, j, jj

	do 12 j=1,n
	s=0.
	if(w(j).ne.0.)then
	do 11 i=1,m
	s=s+u(i,j)*b(i)
	11        continue
	s=s/w(j)
	endif
	tmp(j)=s
	12    continue
	do 14 j=1,n
	s=0.
	do 13 jj=1,n
	s=s+v(j,jj)*tmp(jj)
	13      continue
	x(j)=s
	14    continue

	return

end subroutine svbksb

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine lfit(x,y,sig,ndata,a,ma,lista,mfit,covar,ncvm,chisq,funcs)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! from numerical recipes

	integer, parameter :: mmax=50
	integer :: ndata, mfit,lista(mfit)

	integer :: ma, mp, np, ncvm
	real(kind=dkind):: x(ndata),y(ndata),sig(ndata),a(ma),covar(ncvm,ncvm),beta(mmax),afunc(mmax)
	!v(np,np),u(mp,np),w(np),b(nmax),afunc(mmax)
	real(kind=dkind) :: wmax, chisq, sum, tmp, thresh, s, f, sig2i, ym, wt

!	dimension x(ndata),y(ndata),sig(ndata),a(ma),lista(mfit),covar(ncvm,ncvm),beta(mmax),afunc(mmax)

	external funcs

	integer :: i, j, k, kk, nf, ihit

	kk=mfit+1
	do 12 j=1,ma
	ihit=0
	do 11 k=1,mfit
	if (lista(k).eq.j) ihit=ihit+1
	11      continue
	if (ihit.eq.0) then
	lista(kk)=j
	kk=kk+1
	else if (ihit.gt.1) then
	pause 'improper set in lista'
	endif
	12    continue
	if (kk.ne.(ma+1)) then
		 pause 'improper set in lista'
	endif
	do 14 j=1,mfit
	do 13 k=1,mfit
	covar(j,k)=0.
	13      continue
	beta(j)=0.
	14    continue
	do 18 i=1,ndata
	call funcs(x(i),afunc,ma)
	ym=y(i)
	if(mfit.lt.ma) then
	do 15 j=mfit+1,ma
	ym=ym-a(lista(j))*afunc(lista(j))
	15        continue
	endif
	sig2i=1./sig(i)**2
	do 17 j=1,mfit
	wt=afunc(lista(j))*sig2i
	do 16 k=1,j
	covar(j,k)=covar(j,k)+wt*afunc(lista(k))
	16        continue
	beta(j)=beta(j)+ym*wt
	17      continue
	18    continue
	if (mfit.gt.1) then
	do 21 j=2,mfit
	do 19 k=1,j-1
	covar(k,j)=covar(j,k)
	19        continue
	21      continue
	endif
	call gaussj(covar,mfit,ncvm,beta,1,1)
	do 22 j=1,mfit
	a(lista(j))=beta(j)
	22    continue
	chisq=0.
	do 24 i=1,ndata
	call funcs(x(i),afunc,ma)
	sum=0.
	do 23 j=1,ma
	sum=sum+a(j)*afunc(j)
	23      continue
	chisq=chisq+((y(i)-sum)/sig(i))**2
	24    continue
	!     call covsrt(covar,ncvm,ma,lista,mfit)
	return

end subroutine lfit

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine gaussj(a,n,np,b,m,mp)
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! from numerical recipes

	integer , parameter :: nmax=50
	integer :: n, no, m, mp, np
	real(kind=dkind) ::  a(np,np),b(np,mp)
	integer :: ipiv(nmax),indxr(nmax),indxc(nmax)

	real(kind=dkind) :: big, dum, pivinv

	integer :: i, j, k, irow, icol, l, ll

	do 11 j=1,n
	ipiv(j)=0
	11    continue
	do 22 i=1,n
	big=0.d0
	do 13 j=1,n
	if(ipiv(j).ne.1)then
	do 12 k=1,n
	if (ipiv(k).eq.0) then
	if (abs(a(j,k)).ge.big)then
	big=abs(a(j,k))
	irow=j
	icol=k
	endif
	else if (ipiv(k).gt.1) then
	pause 'singular matrix'
	endif
	12          continue
	endif
	13      continue
	ipiv(icol)=ipiv(icol)+1
	if (irow.ne.icol) then
	do 14 l=1,n
	dum=a(irow,l)
	a(irow,l)=a(icol,l)
	a(icol,l)=dum
	14        continue
	do 15 l=1,m
	dum=b(irow,l)
	b(irow,l)=b(icol,l)
	b(icol,l)=dum
	15        continue
	endif
	indxr(i)=irow
	indxc(i)=icol
	if (a(icol,icol).eq.0.) pause 'singular matrix.'
	pivinv=1./a(icol,icol)
	a(icol,icol)=1.
	do 16 l=1,n
	a(icol,l)=a(icol,l)*pivinv
	16      continue
	do 17 l=1,m
	b(icol,l)=b(icol,l)*pivinv
	17      continue
	do 21 ll=1,n
	if(ll.ne.icol)then
	dum=a(ll,icol)
	a(ll,icol)=0.
	do 18 l=1,n
	a(ll,l)=a(ll,l)-a(icol,l)*dum
	18          continue
	do 19 l=1,m
	b(ll,l)=b(ll,l)-b(icol,l)*dum
	19          continue
	endif
	21      continue
	22    continue
	do 24 l=n,1,-1
	if(indxr(l).ne.indxc(l))then
	do 23 k=1,n
	dum=a(k,indxr(l))
	a(k,indxr(l))=a(k,indxc(l))
	a(k,indxc(l))=dum
	23        continue
	endif
	24    continue

	return

end subroutine gaussj



!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
subroutine checkJ()
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
! for debugging

	use mod_plasma, only : nx, nz, dx, dz, dx_a, dz_a, x_coord, z_coord, sort_grid, j_phi

	implicit none

	real (kind=dkind) :: rhs_loc
	real(kind=dkind) :: term_loc
	real (kind=dkind) :: dA
	real (kind=dkind) :: bigR, bigZ

	integer :: i,j

!	real(kind=dkind) :: R_l, R_r, angle
!	real(kind=dkind) :: psi_integral(2)

	! first, get the current and the cross section area

	rhs_loc = 0.d0

!	dA = dx*dz

	do j=1, nz
		do i=1, nx

		    if(sort_grid(i,j,0)>=1) then
			! may modify to >1 only

				dA = 0.25d0*(dx_a(i) + dx_a(i-1)) * (dz_a(j) + dz_a(j-1))

				term_loc = j_phi(i,j) * dA
				rhs_loc = rhs_loc + term_loc

			endif

		enddo
	enddo

	rhs_loc = -rhs_loc

	continue

	return

end subroutine checkJ






end module mod_equation_system
